/*
Copyright (C) 2012 Taisuke Kobayashi
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
import java.awt.image.BufferedImage;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import javax.imageio.ImageIO;

public class BaseMulticast extends Multicast {

  static final int SERVER_SEND_IMAGE = 0, CLIENT_SEND_IMAGE = 1, REQUEST_REPAINT = 2;
  static final int SEND_IMAGE_START_FLAG = -1, SEND_IMAGE_END_FLAG = -2;

  static final int SEND_IMAGE_RAW = 0, SEND_IMAGE_PNG = 1;
  static final int PACKET_SIZE = 256;
  protected Snapshot snapshot;
  protected String serverId;
  protected int sendImageMode = SEND_IMAGE_RAW;
  protected BufferedImage netImage = null;
  protected byte[] netImageBuf = null;
  protected int saveCounter = SEND_IMAGE_END_FLAG, saveLen = 0;
  protected int[] packetBuf = null;
  protected int[] imageBuf = null;
  protected int imageWidth = 0, imageHeight = 0;

  public BaseMulticast(Snapshot _snapshot) throws Exception{
    super(_snapshot.ini.getMulticastIp(), _snapshot.ini.getMulticastPort());
    snapshot = _snapshot;
    serverId = snapshot.ini.getServerId();
    try{
      sendImageMode = Integer.parseInt(snapshot.ini.getSendImageMode());
    }catch(Exception e){}
    imageWidth = (int)snapshot.windowSize.getWidth();
    imageHeight = (int)snapshot.windowSize.getHeight();
    packetBuf = new int[PACKET_SIZE];
    imageBuf = new int[imageWidth * imageHeight];
  }

  public void setSendImageMode(int mode){
    if((sendImageMode < SEND_IMAGE_RAW) || (sendImageMode > SEND_IMAGE_PNG))
      return;
    sendImageMode = mode;
  }

  public void receiveImage(BaseCanvas bc) throws Exception{
    int index = getStartIndex();
    switch(getInt(index)){
    default:
    case SEND_IMAGE_RAW:
      receiveImageRaw(bc, index +4);
      break;
    case SEND_IMAGE_PNG:
      receiveImagePng(bc, index +4);
      break;
    }
  }

  public void sendImage(int cmd, BufferedImage bi) throws Exception{
    switch(sendImageMode){
    default:
    case SEND_IMAGE_RAW:
      sendImageRaw(cmd, bi);
      break;
    case SEND_IMAGE_PNG:
      sendImagePng(cmd, bi);
      break;
    }
  }

  public void receiveImageRaw(BaseCanvas bc, int index) throws Exception{
    try{
      int x = getInt(index); index += 4;
      int y = getInt(index); index += 4;
      int dx = getInt(index); index += 4;
      if(dx > PACKET_SIZE)
        dx = PACKET_SIZE;
      for(int j = 0; j < dx; j++)
        packetBuf[j] = getInt(j *4 + index);
      bc.setNetImageRaw(x, y, dx, 1, packetBuf);
    }catch(Exception e){
      C.throwException("BaseMulticast: receiveImage:", e);
    }
  }

  public void sendImageRaw(int cmd, BufferedImage bi) throws Exception{
    try{
      int w = imageWidth, h = imageHeight;
      for(int y = 0; y < h; y++){
        for(int x = 0; x < w; x += PACKET_SIZE){
          int dx = w - x;
          if(dx > PACKET_SIZE)
            dx = PACKET_SIZE;
          bi.getRGB(x, y, dx, 1, packetBuf, 0, dx);
          boolean yet = true;
          int di = x + y * w;
          for(int i = 0; i < dx; i++){
            if(yet && (imageBuf[di + i] != packetBuf[i])){
              int index = setCmdAndServerId(cmd);
              setInt(SEND_IMAGE_RAW, index); index += 4;
              setInt(x, index); index += 4;
              setInt(y, index); index += 4;
              setInt(dx, index); index += 4;
              for(int j = 0; j < dx; j++)
                setInt(packetBuf[j], j *4 + index);
              send();
              yet = false;
            }
            imageBuf[di + i] = packetBuf[i];
          }
        }
      }
      setCmdAndServerId(REQUEST_REPAINT);
      send();
      C.sleep(1);
    }catch(Exception e){
      C.throwException("BaseMulticast: sendImageRaw:", e);
    }
  }

  public void receiveImagePng(BaseCanvas bc, int index) throws Exception{
    try{
      int i1 = 0, i2 = 0;
      i1 = getInt(index); index += 4;
      switch(i1){
      case SEND_IMAGE_START_FLAG:
        i2 = getInt(index);
        netImageBuf = new byte[i2];
        saveCounter = i1;
        saveLen = 0;
        break;
      case SEND_IMAGE_END_FLAG:
        if(saveCounter == SEND_IMAGE_END_FLAG)
          break;
        i2 = getInt(index);
        if(i2 == saveCounter +1){
          netImage = ba2bi(netImageBuf);
          bc.copyNetImage(netImage);
          netImageBuf = null;
          netImage = null;
        }else{
          C.e("BaseMulticast: receiveImage: i2 != saveCounter +1 " + i2 + " " + saveCounter);
          saveCounter = SEND_IMAGE_END_FLAG;
        }
        break;
      default:
        if(saveCounter == SEND_IMAGE_END_FLAG)
          break;
        if(i1 != saveCounter +1){
          C.e("BaseMulticast: receiveImage: i1 != saveCounter +1 " + i1 + " " + saveCounter);
          saveCounter = SEND_IMAGE_END_FLAG;
        }else{
          i2 = getInt(index); index += 4;
          System.arraycopy(rcvBuf, index, netImageBuf, saveLen, i2);
          saveLen += i2;
          saveCounter = i1;
        }
        break;
      }
    }catch(Exception e){
      C.throwException("BaseMulticast: receiveImagePng:", e);
    }
  }

  public void sendImagePng(int cmd, BufferedImage bi) throws Exception{
    try{
      int sleepMsec = 1;
      byte[] ba = bi2ba(bi);
      int index = setCmdAndServerId(cmd); // start packet
      int len1 = ba.length, len2 = BUF_SIZE - index -12, len3 = 0, n = (len1 + len2) / len2;
      setInt(SEND_IMAGE_PNG, index); index += 4;
      setInt(SEND_IMAGE_START_FLAG, index); index += 4;
      setInt(len1, index);
      send();
      C.sleep(sleepMsec);
      int i = 0;
      for(i = 0; i < n; i++){
        index = setCmdAndServerId(cmd);
        setInt(SEND_IMAGE_PNG, index); index += 4;
        setInt(i, index); index += 4;
        if(len1 - len3 <= len2){
          int len = len1 - len3;
          setInt(len, index); index += 4;
          System.arraycopy(ba, len3, sndBuf, index, len);
          send();
          C.sleep(sleepMsec);
          break;
        }else{
          setInt(len2, index); index += 4;
          System.arraycopy(ba, len3, sndBuf, index, len2);
          len3 += len2;
          send();
          C.sleep(sleepMsec);
        }
      }
      index = setCmdAndServerId(cmd); // end packet
      setInt(SEND_IMAGE_PNG, index); index += 4;
      setInt(SEND_IMAGE_END_FLAG, index); index += 4;
      setInt(i +1, index);
      send();
      C.sleep(sleepMsec);
    }catch(Exception e){
      C.throwException("BaseMulticast: sendImagePng:", e);
    }
  }

  public byte[] bi2ba(BufferedImage bi) throws Exception{
    ByteArrayOutputStream ba = new ByteArrayOutputStream();
    BufferedOutputStream bs = new BufferedOutputStream(ba);
    bi.flush();
    ImageIO.write(bi, Snapshot.IMAGE_TYPE, bs);
    bs.flush();
    bs.close();
    return ba.toByteArray();
  }

  public BufferedImage ba2bi(byte[] ba) throws Exception{
    return ImageIO.read(new ByteArrayInputStream(ba));
  }

  public String getServerId(){ return getString(8, getInt(4)); }

  public int getStartIndex(){ return getInt(4) +8; }

  public int getCmd(){ return getInt(0); }

  public int setCmdAndServerId(int cmd){
    int len = serverId.length();
    setInt(cmd, 0);
    setInt(len, 4);
    setString(serverId, 8);
    return len +8;
  }

  public String getString(int index, int len){
    return new String(rcvBuf, index, len);
  }

  public void setString(String s, int index){
    System.arraycopy(s.getBytes(), 0, sndBuf, index, s.length());
  }

  public int getInt(int index){
    int r = 0;
    for(int i = 0; i < 4; i++){
      r |= rcvBuf[index + i] & 0xff;
      if(i < 3)
        r <<= 8;
    }
    return r;
  }

  public void setInt(int v, int index){
    for(int i = 0; i < 4; i++){
      sndBuf[index +3 - i] = (byte)(v & 0xff);
      v >>= 8;
    }
  }

}
